#! /usr/bin/env python
# -*- coding: utf-8 -*-
import os
import json
from msal import application
import requests
import msal
from loguru import logger
import sys
from urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter
from typing import List

from .exceptions import handle_http_errors, ClientCredentialsError, AcquireTokenException
from .utils import timed


DEFAULT_TIMEOUT = 300  # seconds
DEFAULT_SESSION_LIFETIME = 300  # seconds
DEFAULT_AUTHORITY = os.environ.get('GRAPH_API_AUTHORITY', '')
DEFAULT_CLIENT_ID = os.environ.get('GRAPH_API_CLIENT_ID', '')
DEFAULT_CLIENT_SECRET = os.environ.get('GRAPH_API_CLIENT_SECRET', '')


logger.disable('msgraph')


class MicrosoftGraphClient():
    def __init__(self, authority: str = None, client_id: str = None, client_secret: str = None):

        self.authority = authority or DEFAULT_AUTHORITY
        self.client_id = client_id or DEFAULT_CLIENT_ID
        self.client_secret = client_secret or DEFAULT_CLIENT_SECRET
        self.scope = ['https://graph.microsoft.com/.default']
        self.grant_type = 'client_credentials'

        try:
            self.app = msal.ConfidentialClientApplication(
                self.client_id, authority=self.authority,
                client_credential=self.client_secret
            )
        except ValueError as e:
            logger.exception(e)
            raise ClientCredentialsError(e)

        # init http session
        self.http = requests.Session()
        retry_strategy = Retry(
            total=3,
            status_forcelist=[429, 500, 502, 503, 504],
            allowed_methods=frozenset({"HEAD", "GET", "PUT",
                                       "POST", "PATCH", "DELETE", "OPTIONS"}),
            backoff_factor=1
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        self.http.mount("https://", adapter)
        self.http.mount("http://", adapter)

    @timed
    @logger.catch(reraise=True)
    def get_token(self):
        result = None
        result = self.app.acquire_token_silent(self.scope, account=None)

        if not result:
            result = self.app.acquire_token_for_client(scopes=self.scope)

        if 'access_token' in result:
            return result['access_token']
        else:
            raise AcquireTokenException(
                result.get('error'),
                result.get('error_description'),
                result.get('description')
            )

    def get_workbook(self, drive_id: str, book_id: str):
        return Workbook(client=self, drive_id=drive_id, book_id=book_id)


class Workbook():
    """ClientCredentialsError
    Позволяет читать и писать в книгу Excel.
    Все операции производятся с открытыми сессиями
    """

    def __init__(self, client: MicrosoftGraphClient, drive_id: str, book_id: str):
        self.client = client
        self.drive_id = drive_id
        self.book_id = book_id

    @logger.catch(reraise=True)
    def create_session(self, persist: bool = True):
        """
        Создает новый сеанс книги
        Params: 
        - persist(bool): True - постоянный сеанс, False - временный сеанс
        Return: 
        - session_id(str): Возвращает идентификатор сессии
        """
        url = f"https://graph.microsoft.com/v1.0/drives/{self.drive_id}/items/{self.book_id}/workbook/createSession"
        headers = headers = {
            "Authorization": f"Bearer {self.client.get_token()}",
            "Content-type": "application/json"
        }
        payload = {
            "persistChanges": persist
        }

        data = json.dumps(payload)

        response = handle_http_errors(
            self.client.http.post(url, headers=headers, data=data))

        session_id = response.json()['id']
        return Session(self, session_id, persist)

    @logger.catch(reraise=True)
    def get_range_values(self, page_name: str, address: str, session=None) -> List[List]:
        """
        Возвращает значения из диапазона страницы
        Params:
        - page_name(str): Имя страницы в книге
        - address(str): Адресс диапазона ячеек на странице в виде - {A1:B2}
        Return:
        - values: list[list[str, int]] - список строк запрашиваемого диапазона. Каждая строка - список значений ячеек
        """

        url = f"https://graph.microsoft.com/v1.0/drives/{self.drive_id}/items/{self.book_id}/workbook/worksheets/{page_name}/range(address='{address}')"
        headers = {
            "Authorization": f"Bearer {self.client.get_token()}",
            "Content-type": "application/json",
        }
        if session:
            headers.update({'workbook-session-id': session.get_id()})

        response = handle_http_errors(
            self.client.http.get(url, headers=headers))

        return response.json()['values']

    @logger.catch(reraise=True)
    def update_range_values(self, page_name: str, address: str, values: List[List], session=None) -> List[List]:
        """
        Возвращает значения из диапазона страницы
        Params:
        - page_name(str): Имя страницы в книге
        - address(str): Адресс диапазона ячеек на странице в виде - {A1:B2}
        - values: list[list[str, int]] - список строк запрашиваемого диапазона. Каждая строка - список значений ячеек
        Return:
        - updated_values: list[list[str, int]] - список строк обновленного диапазона. Каждая строка - список значений ячеек
        """
        url = f"https://graph.microsoft.com/v1.0/drives/{self.drive_id}/items/{self.book_id}/workbook/worksheets/{page_name}/range(address='{address}')"
        headers = {
            "Authorization": f"Bearer {self.client.get_token()}",
            "Content-type": "application/json",
        }
        payload = {
            'values': values
        }

        if session:
            headers.update({'workbook-session-id': session.get_id()})

        response = handle_http_errors(self.client.http.patch(
            url, headers=headers, data=json.dumps(payload)))

        return response.json()['values']

    @logger.catch(reraise=True)
    def clear_range(self, page_name: str, address:str, session=None) -> None:
        """
        Очищает значения в диапазоне страницы
        Params:
        - page_name(str): Имя страницы в книге
        - address(str): Адресс диапазона ячеек на странице в виде - {A1:B2}
        """
        url = f"https://graph.microsoft.com/v1.0/drives/{self.drive_id}/items/{self.book_id}/workbook/worksheets/{page_name}/range(address='{address}')/format/fill/clear"
        headers = {
            "Authorization": f"Bearer {self.client.get_token()}",
            "Content-type": "application/json",
        }
        if session:
            headers.update({'workbook-session-id': session.get_id()})

        handle_http_errors(
            self.client.http.post(url, headers=headers))

    @logger.catch(reraise=True)
    def calculate(self, calculation_type: str) -> None:
        """
        см. https://docs.microsoft.com/ru-ru/graph/api/workbookapplication-calculate?view=graph-rest-1.0
        Пересчитывает данные во всех открытых в текущий момент книгах Excel.
        Params:
            сalculation_type(str): Определяет тип расчета, который нужно использовать. Возможные значения: Recalculate, Full, FullRebuild.
        """

        url = f"https://graph.microsoft.com/v1.0/drives/{self.drive_id}/items/{self.book_id}/workbook/application/calculate"
        headers = {
            "Authorization": f"Bearer {self.client.get_token()}",
            "Content-type": "application/json",
        }

        payload = {
            "calculationType": calculation_type
        }

        handle_http_errors(self.client.http.post(
            url, headers=headers, data=json.dumps(payload)
        ))


class Session():
    def __init__(self, workbook: Workbook, session_id: str, persist: bool):
        self.workbook = workbook
        self.session_id = session_id
        self.persist = persist

    def __enter__(self):
        self.update()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.close()

    def get_id(self):
        return self.session_id

    @timed
    @logger.catch(reraise=True)
    def update(self):
        """
        Обновляет сеанс книги
        Params:
        - session-id - идентификатор сессии которую нужно обновить
        """
        url = f"https://graph.microsoft.com/v1.0/drives/{self.workbook.drive_id}/items/{self.workbook.book_id}/workbook/refreshSession"
        headers = {
            "Authorization": f"Bearer {self.workbook.client.get_token()}",
            "Content-type": "application/json",
            "workbook-session-id": self.get_id(),
        }

        handle_http_errors(
            self.workbook.client.http.post(url, headers=headers))

    @timed
    @logger.catch(reraise=True)
    def close(self):
        """
        Закрывает текущий сеанс книги
        """
        url = f"https://graph.microsoft.com/v1.0/drives/{self.workbook.drive_id}/items/{self.workbook.book_id}/workbook/closeSession"
        headers = {
            "Authorization": f"Bearer {self.workbook.client.get_token()}",
            "Content-type": "application/json",
            "workbook-session-id": self.get_id(),
        }

        handle_http_errors(
            self.workbook.client.http.post(url, headers=headers))

        self.session_id = None

    def get_range_values(self, page_name: str, address: str) -> List[List]:
        return self.workbook.get_range_values(page_name=page_name, address=address, session=self)

    def update_range_values(self, page_name: str, address: str, values: List[List]) -> List[List]:
        return self.workbook.update_range_values(page_name=page_name, address=address, values=values, session=self)
