from loguru import logger
from configparser import ConfigParser
import random
import sys
from unittest import TestCase

from msgraph.client import MicrosoftGraphClient, Workbook
from msgraph.exceptions import AcquireTokenException, ClientCredentialsError, \
    InvalidRequestException, ItemNotFoundException

config = {
    "handlers": [
        {"sink": sys.stdout, "level": "WARNING", "format": "{time} - {message}"},
        {"sink": "tests/logs/file.log", "level": "DEBUG", "serialize": True},
    ],
}
logger.configure(**config)
logger.enable('msgraph')


class TestClient(TestCase):
    def __init__(self, *args, **kwargs):
        config = ConfigParser()
        config.read('tests/config.ini')

        self.authority = config.get('test_graph_api', 'authority')
        self.client_id = config.get('test_graph_api', 'client_id')
        self.client_secret = config.get('test_graph_api', 'client_secret')

        super(TestClient, self).__init__(*args, **kwargs)

    def test_default(self):
        """
        Must past default test
        """
        client = MicrosoftGraphClient(
            self.authority, self.client_id, self.client_secret)

    def test_wrong_authority(self):
        """Wrong authority must rise ClientCredentialsError"""
        with self.assertRaises(ClientCredentialsError) as e:
            client = MicrosoftGraphClient(
                authority='https://login.microsoftonline.com/me',
                client_id=self.client_id,
                client_secret=self.client_secret).get_token()

    def test_wrong_client_id(self):
        """Wrong client id must rise AcquireTokenException"""
        with self.assertRaises(AcquireTokenException):
            MicrosoftGraphClient(self.authority,
                                 client_id="1", client_secret=self.client_secret).get_token()

    def test_wrong_client_secret(self):
        """Wrong client secret must rise AcquireTokenException"""
        with self.assertRaises(AcquireTokenException):
            MicrosoftGraphClient(
                self.authority, client_id=self.client_id, client_secret=1).get_token()


class TestWorkbook(TestCase):
    def __init__(self, *args, **kwargs):

        config = ConfigParser()
        config.read('tests/config.ini')

        self.authority = config.get('test_graph_api', 'authority')
        self.client_id = config.get('test_graph_api', 'client_id')
        self.client_secret = config.get('test_graph_api', 'client_secret')
        self.drive_id = config.get('test_graph_api', 'drive_id')
        self.book_id = config.get('test_graph_api', 'book_id')
        self.input_page_name = config.get('test_graph_api', 'input_page_name')
        self.input_address = config.get('test_graph_api', 'input_address')
        self.output_page_name = config.get(
            'test_graph_api', 'output_page_name')
        self.output_address = config.get('test_graph_api', 'output_address')

        self.manual_calculation_book_id = config.get(
            'test_graph_api', 'manual_calculation_book_id')

        self.client = MicrosoftGraphClient(authority=self.authority,
                                           client_id=self.client_id,
                                           client_secret=self.client_secret)

        super(TestWorkbook, self).__init__(*args, **kwargs)

    def test_get_workbook_default(self):
        '''
        Все должно просто сработать
        '''
        workbook = self.client.get_workbook(self.drive_id, self.book_id)
        self.assertIsInstance(workbook, Workbook)
        self.assertIsNotNone(workbook)

    def test_get_workbook_wrong_book_id(self):
        '''
        Не верный book_id должен вернуть ItemNotFound
        '''
        with self.assertRaises(ItemNotFoundException):
            self.client.get_workbook(self.drive_id, '123').create_session()

    def test_page_not_found(self):
        """
        Если страница не существует в книге, то необходимо вернуть ItemNotFoundException
        """
        workbook = self.client.get_workbook(self.drive_id, self.book_id)

        with self.assertRaises(ItemNotFoundException):
            workbook.update_range_values(
                "no such page", "A1:B1", [['1', '2']]
            )

        with self.assertRaises(ItemNotFoundException):
            workbook.get_range_values(
                'no such page', "A1:B1"
            )

    def test_update_range_values_without_session(self):
        input = [[1, 2]]
        workbook = self.client.get_workbook(self.drive_id, self.book_id)
        result = workbook.update_range_values(
            self.input_page_name, self.input_address, input)

        self.assertEqual(result, input)

    def test_get_range_values_without_session(self):
        workbook = self.client.get_workbook(self.drive_id, self.book_id)
        result = workbook.get_range_values(self.output_page_name, 'A3:B3')

        self.assertEqual(result, [[100, 'B']])

    def test_update_and_get_range_values_with_session(self):
        workbook = self.client.get_workbook(self.drive_id, self.book_id)
        with workbook.create_session() as session:
            input = [random.randint(0, 1000), random.randint(0, 1000)]
            session.update_range_values(
                self.input_page_name, self.input_address, [input])
            result = session.get_range_values(
                self.output_page_name, 'A2')
            self.assertEqual(result[0][0], sum(input))

    def test_calculate(self):
        workbook = self.client.get_workbook(
            self.drive_id, self.manual_calculation_book_id)
        with workbook.create_session() as session:
            input = [random.randint(0, 1000), random.randint(0, 1000)]
            session.update_range_values(
                "Input", "A2:B2", [input]
            )
            workbook.calculate(calculation_type="Recalculate")
            result = session.get_range_values(
                "Output", "A2"
            )
            self.assertEqual(result[0][0], sum(input))
